//
//  CommunityItemView.swift
//  SampleApp
//
//  Created by Nontapat Siengsanor on 11/8/2564 BE.
//  Copyright © 2564 BE David Zhang. All rights reserved.
//

import SwiftUI

struct CommunityItemView: View {
    typealias Action = (() -> Void)?
    
    var model: CommunityModel
    var deleteAction: (() -> Void)?
    var leaveAction: (() -> Void)?
    var detailsAction: (() -> Void)?
    var joinAction: (() -> Void)?
    var updateAction: (() -> Void)?
    var membershipAction: (() -> Void)?
    var feedAction: (() -> Void)?
    var viewStoryAction: (() -> Void)?
    var realTimeEventAction: (() -> Void)?
    var checkPermissionAction: (() -> Void)?
    var copyCommunityIdAction: (() -> Void)?
    
    @State var showAction: Bool = false
    @State private var hasEditCommunityUserPermission: Bool = false
    @State private var hasEditCommunityPermission: Bool = false
    
    init(model: CommunityModel,
         joinAction: Action,
         leaveAction: Action,
         deleteAction: Action,
         detailsAction: Action,
         updateAction: Action,
         membershipAction: Action,
         feedAction: Action,
         viewStoryAction: Action,
         realTimeEventAction: Action,
         checkPermissionAction: Action,
         copyCommunityIdAction: Action) {
        self.model = model
        self.joinAction = joinAction
        self.leaveAction = leaveAction
        self.deleteAction = deleteAction
        self.detailsAction = detailsAction
        self.updateAction = updateAction
        self.membershipAction = membershipAction
        self.feedAction = feedAction
        self.viewStoryAction = viewStoryAction
        self.realTimeEventAction = realTimeEventAction
        self.checkPermissionAction = checkPermissionAction
        self.copyCommunityIdAction = copyCommunityIdAction
    }
    
    var body: some View {
        VStack {
            HStack {
                VStack(alignment: .leading) {
                    Text("\(model.displayName)")
                        .font(.title)
                        .fontWeight(.semibold)
                        .foregroundColor(.white)
                        .padding(.top, 12)

                    if !model.description.isEmpty {
                        Text("\(model.description.capitalized)")
                            .foregroundColor(Color.white.opacity(0.9))
                            .padding(.top, 4)
                            .font(.body)
                    }
                }
                .padding(.horizontal, 20)
                .padding(.top, 10)

                Spacer()

                Image(systemName: "ellipsis.circle")
                    .resizable()
                    .foregroundColor(.white)
                    .frame(width: 24, height: 24)
                    .padding()
                    .onTapGesture {
                        fetchPermission {
                            showAction.toggle()
                        }
                    }
                .actionSheet(isPresented: $showAction, content: {
                    return ActionSheet(title: Text("Choose Actions"), message: Text("What do you want to do?"), buttons: getAllActions())
                })
            }

            VStack(spacing: 5) {

                let data = model.getPreviewData()

                ForEach(0..<data.count, id: \.self) { index in
                    let item = data[index]
                    CardTextView(key: item.0, value: item.1)
                }

            }
            .padding([.top, .bottom], 16)
        }
        .background(model.isJoined ? Color("NeonGreen") : Color("PastelBlue"))
        .cornerRadius(20)
        .shadow(color: Color.gray.opacity(0.6), radius: 8, x: 0, y: 0)
        .padding([.top, .bottom], 8)
    }
    
    func fetchPermission(completion: (() -> Void)?) {
        let group = DispatchGroup()
            
        group.enter()
        AmityManager.shared.client?.hasPermission(.editCommunityUser, forCommunity: model.communityId) { permission in
            hasEditCommunityUserPermission = permission
            group.leave()
        }
        
        group.enter()
        AmityManager.shared.client?.hasPermission(.editCommunity, forCommunity: model.communityId) { permission in
            hasEditCommunityPermission = permission
            group.leave()
        }
        
        group.notify(queue: .main) {
            completion?()
        }
    }
    
    func getAllActions() -> [ActionSheet.Button] {
        var actions = [ActionSheet.Button]()
        
        if model.isJoined {
            actions.append(.default(Text("View Community Details"), action: {
                detailsAction?()
            }))
            
            actions.append(.default(Text("View Community Feed"), action: {
                feedAction?()
            }))
            
            actions.append(.default(Text("View Community Story"), action: {
                viewStoryAction?()
            }))            
            
            actions.append(.default(Text("Leave Community"), action: {
                leaveAction?()
            }))
            
            actions.append(.default(Text("Real Time Event"), action: {
                realTimeEventAction?()
            }))
            
            actions.append(.default(Text("Check permission"), action: {
                checkPermissionAction?()
            }))
            
            // Actions related to community permissions
            if hasEditCommunityUserPermission {
                actions.append(.default(Text("Add/Remove Members"), action: {
                    membershipAction?()
                }))
            }
            if hasEditCommunityPermission {
                
                actions.append(.default(Text("Update Community"), action: {
                    updateAction?()
                }))
                
                actions.append(.default(Text("Delete Community"), action: {
                    deleteAction?()
                }))
            }
            actions.append(.default(Text("Copy Community Id"), action: {
                copyCommunityIdAction?()                
            }))
        } else {
            actions.append(.default(Text("Join Community"), action: {
                joinAction?()
            }))
        }
        
        // Default dismiss action
        actions.append(.cancel(Text("Dismiss")))
        
        return actions
    }
}

struct CardTextView: View {
    
    let key: String
    let value: String
    var lineLimit:Int? = 1
    
    var body: some View {
        HStack(alignment: .top) {
            Text(key)
                .padding(.leading, 20)
                .foregroundColor(Color.white)
                .font(.system(size: 15))
            
            Text(value)
                .foregroundColor(Color.white.opacity(0.8))
                .font(.system(size: 15))
                .lineLimit(lineLimit)
            
            Spacer()
        }
    }
}
