//
//  CommunityListViewModel.swift
//  SampleApp
//
//  Created by Michael Abadi on 13/07/20.
//  Copyright © 2020 David Zhang. All rights reserved.
//

import Foundation
import AmitySDK

struct CommunityModel: Identifiable, Equatable {

    let communityObject: AmityCommunity
    
    static func == (lhs: CommunityModel, rhs: CommunityModel) -> Bool {
        return lhs.id == rhs.id
    }
    
    var id: String {
        return communityId
    }
    let communityId: String
    let description: String
    let displayName: String
    let isPublic: Bool
    let isOfficial: Bool
    var isJoined: Bool = false
    let channelId: String
    let postsCount: Int
    var membersCount: Int
    let createdAt: Date
    let metadata: [String: Any]?
    let userId: String
    let tags: [String]
    let categoryIds: [String]?
    let isPostReview: Bool
    let postSettings: AmityCommunityPostSettings
    let storySettings: AmityCommunityStorySettings
    
    var isCreator: Bool {
        return AmityManager.shared.client?.currentUserId == userId
    }
    
    init(community: AmityCommunity) {
        communityObject = community
        communityId = community.communityId
        description = community.communityDescription
        displayName = community.displayName
        isPublic = community.isPublic
        isOfficial = community.isOfficial
        isJoined = community.isJoined
        channelId = community.channelId
        postsCount = Int(community.postsCount)
        membersCount = Int(community.membersCount)
        createdAt = community.createdAt
        metadata = community.metadata
        userId = community.userId
        tags = community.tags ?? []
        categoryIds = community.categoryIds
        isPostReview = community.isPostReviewEnabled
        postSettings = community.postSettings
        storySettings = community.storySettings
    }
    
    func getCommunityObject() -> AmityCommunity {
        return communityObject
    }
    
    func getPreviewData() -> [(key: String, value: String)] {
        return [
            ("Community Id:", communityId),
            ("Display Name:", displayName),
            ("Is Public:", isPublic ? "YES" : "NO"),
            ("MetaData", metadata?.description ?? ""),
            ("Created At:", "\(createdAt)"),
            ("Categories", communityObject.categories.map{ $0.name }.joined(separator: ",")),
            ("Categories Count:", "\(categoryIds?.count ?? 0)"),
            ("Members Count:", "\(membersCount)"),
            ("Post Count:", "\(postsCount)"),
            ("Post Published Count:", "\(getCommunityObject().getPostCount(feedType: .published))"),
            ("Post Reviewing Count:", "\(getCommunityObject().getPostCount(feedType: .reviewing))"),
            ("Only Admin Can Post", communityObject.onlyAdminCanPost ? "YES" : "NO"),
            ("Post Review Enabled", isPostReview ? "YES" : "NO"),
            ("Post Settings:", communityObject.postSettings.title),
            ("Story Settings:", CommunityStorySettings.convertFromAmitySettings(communityObject.storySettings).title)
        ]
    }
}

enum CommunityType: String {
    case normal = "Default"
    case trending = "Trending"
    case recommended = "Recommended"
}

// Concrete implementor
class CommunityListViewModel: ObservableObject {
    
    var type: CommunityType
    var pageTitle: String
    
    private let communityRepository = AmityCommunityRepository(client: AmityManager.shared.client!)
    
    // Get notified when this collection changes.
    // Note: This does not observe changes to snapshot inside this collection.
    @Published var communityCollection: AmityCollection<AmityCommunity>
        
    var searchKeyword: String = ""
    
    private var filter: AmityCommunityQueryFilter = .all
    private var sort: AmityCommunitySortOption = .lastCreated
    
    var currentSortOption: AmityCommunityCategoriesSortOption = .displayName
    var categoryId = ""
    var shouldIncludeDeletedCommunity = false
    var shouldIncludeDeletedCategories = false
    
    init(type: CommunityType) {
        self.type = type
        self.pageTitle = type.rawValue
        
        switch type {
        case .normal:
            let queryOptions = AmityCommunityQueryOptions(displayName: searchKeyword, filter: filter, sortBy: sort, categoryId: categoryId, includeDeleted: shouldIncludeDeletedCommunity)
            communityCollection = communityRepository.getCommunities(with: queryOptions)
        case .recommended:
            communityCollection = communityRepository.getRecommendedCommunities()
            
        case .trending:
            communityCollection = communityRepository.getTrendingCommunities()
        }
    }
    
    func queryCommunity() {
        switch type {
        case .normal:
            let queryOptions = AmityCommunityQueryOptions(displayName: searchKeyword, filter: filter, sortBy: sort, categoryId: categoryId, includeDeleted: shouldIncludeDeletedCommunity)
            communityCollection = communityRepository.getCommunities(with: queryOptions)
        case .recommended:
            communityCollection = communityRepository.getRecommendedCommunities()
        case .trending:
            communityCollection = communityRepository.getTrendingCommunities()
        }
    }
    
    func searchCommunities() {
        self.queryCommunity()
    }
    
    func setSort(_ sort: AmityCommunitySortOption) {
        self.sort = sort
    }
    
    func setFilter(_ filter: AmityCommunityQueryFilter) {
        self.filter = filter
    }
    
    func setIncludeDeleted(_ includeDeleted: Bool) {
        self.shouldIncludeDeletedCommunity = includeDeleted
    }
    
    func setCategoryId(_ categoryId: String) {
        self.categoryId = categoryId
    }
    
    func fetchNextPage() {
        guard communityCollection.hasNext else { return }
        communityCollection.nextPage()
    }
    
    func deleteCommunity(for community: CommunityModel, completion: ((Bool, Error?) -> Void)?) {
        communityRepository.deleteCommunity(withId: community.communityId) { (success, error) in
            completion?(success, error)
        }
    }
    
    func joinCommunity(for community: CommunityModel, completion: ((Bool, Error?) -> Void)?) {
        Task { @MainActor in
            var success: Bool = false
            var err: Error?
            do {
                let result = try await communityRepository.joinCommunity(withId: community.communityId)
                success = result
            } catch {
                err = error
            }
            completion?(success, err)
        }
    }
    
    func leaveCommunity(for community: CommunityModel, completion: ((Bool, Error?) -> Void)?) {
        Task { @MainActor in
            do {
                let result = try await communityRepository.leaveCommunity(withId: community.communityId)
                completion?(result, nil)
            } catch {
                completion?(false, error)
                Log.add(info: error)
            }
        }
    }
}
